/*
 * Please do not edit this file.
 * It was generated using rpcgen.
 */

#include <rpc/types.h>


#define yp_server ((u_long)0x20001003)
#define yp_server_version ((u_long)1)
#define do_yp_frame ((u_long)301)
extern char **do_yp_frame_1();

