/* -*-C-*-

$Id: os2pmcon.h,v 1.6 2003/02/14 18:28:22 cph Exp $

Copyright (c) 1995, 1999 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

/* Resource IDs for OS/2 PM Console Window */

#define ID_PMCON_RESOURCES 2

#define IDM_FILE	1
#define IDM_EDIT	2
#define IDM_OPTIONS	3
#define IDM_HELP	4

#define IDM_EXIT	10

#define IDM_CUT		20
#define IDM_COPY	21
#define IDM_PASTE	22

#define IDM_FONT	30

#define IDM_ABOUT	40

#define IDI_BCH		10
#define IDI_COFFEE	11
#define IDI_CONSES	12
#define IDI_EDWIN	13
#define IDI_ENVIR1	14
#define IDI_GRAPHICS	15
#define IDI_LAMBDA	16
#define IDI_LAMBDA2	17
#define IDI_LIAR1	18
#define IDI_LIAR2	19
#define IDI_LIAR3	20
#define IDI_MINCER	21
#define IDI_SHIELD1	22
#define IDI_SHIELD2	23
#define IDI_SHIELD3	24
#define IDI_SHIELD4	25
