@echo off
rem MIT/GNU Scheme microcode configuration script for Windows NT
rem
rem Copyright 1993 Massachusetts Institute of Technology
rem
rem $Id: setenv.bat,v 1.2 2003/02/14 18:48:12 cph Exp $
rem
set path=%path%;c:\usr\lib\masm\bin
set include=%lib%;c:\usr\lib\masm\include
d:\mstools\setenv d:\mstools
